/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/03/07 14:41:43
 */

import { is } from './utils.js'
import { rad, deg } from './lib/math.js'

function norm(a) {
  return a[0] * a[0] + a[1] * a[1]
}
function normalize(a) {
  let mag = Math.sqrt(norm(a))
  a[0] && (a[0] /= mag)
  a[1] && (a[1] /= mag)
}

export class Matrix {
  a = 1
  b = 0
  c = 0
  d = 1
  e = 0
  f = 0

  constructor(a, b, c, d, e, f) {
    if (b === void 0 && is(a, 'SVGMatrix')) {
      this.a = a.a
      this.b = a.b
      this.c = a.c
      this.d = a.d
      this.e = a.e
      this.f = a.f
      return
    }
    if (a !== void 0) {
      this.a = +a
      this.b = +b
      this.c = +c
      this.d = +d
      this.e = +e
      this.f = +f
    }
  }

  /**
   * Adds the given matrix to existing one
   */
  add(a, b, c, d, e, f) {
    if (a && a instanceof Matrix) {
      return this.add(a.a, a.b, a.c, a.d, a.e, a.f)
    }
    let aNew = a * this.a + b * this.c,
      bNew = a * this.b + b * this.d
    this.e += e * this.a + f * this.c
    this.f += e * this.b + f * this.d
    this.c = c * this.a + d * this.c
    this.d = c * this.b + d * this.d

    this.a = aNew
    this.b = bNew
    return this
  }

  /**
   * Multiplies a passed affine transform to the left: M * this.
   */
  multLeft(a, b, c, d, e, f) {
    if (a && a instanceof Matrix) {
      return this.multLeft(a.a, a.b, a.c, a.d, a.e, a.f)
    }
    let aNew = a * this.a + c * this.b,
      cNew = a * this.c + c * this.d,
      eNew = a * this.e + c * this.f + e
    this.b = b * this.a + d * this.b
    this.d = b * this.c + d * this.d
    this.f = b * this.e + d * this.f + f

    this.a = aNew
    this.c = cNew
    this.e = eNew
    return this
  }

  /**
   * Returns an inverted version of the matrix
   */
  invert() {
    let me = this
    let x = me.a * me.d - me.b * me.c
    return new Matrix(
      me.d / x,
      -me.b / x,
      -me.c / x,
      me.a / x,
      (me.c * me.f - me.d * me.e) / x,
      (me.b * me.e - me.a * me.f) / x
    )
  }
  /**
   * Returns a copy of the matrix
   */
  clone() {
    return new Matrix(this.a, this.b, this.c, this.d, this.e, this.f)
  }
  /**
   * Translate the matrix
   * x (number) horizontal offset distance
   * y (number) vertical offset distance
   */
  translate(x, y) {
    this.e += x * this.a + y * this.c
    this.f += x * this.b + y * this.d
    return this
  }

  /**
   * Rotates the matrix
   * a (number) angle of rotation, in degrees
   * x (number) horizontal origin point from which to rotate
   * y (number) vertical origin point from which to rotate
   */
  rotate(a, x, y) {
    a = rad(a)
    x = x || 0
    y = y || 0
    let cos = +Math.cos(a).toFixed(9),
      sin = +Math.sin(a).toFixed(9)
    this.add(cos, sin, -sin, cos, x, y)
    return this.add(1, 0, 0, 1, -x, -y)
  }
  /**
   * Skews the matrix along the x-axis
   * x (number) Angle to skew along the x-axis (in degrees).
   */
  skewX(x) {
    return this.skew(x, 0)
  }
  /**
   * Skews the matrix along the y-axis
   * y (number) Angle to skew along the y-axis (in degrees).
   */
  skewY(y) {
    return this.skew(0, y)
  }
  /**
   * Skews the matrix
   * y (number) Angle to skew along the y-axis (in degrees).
   * x (number) Angle to skew along the x-axis (in degrees).
   */
  skew(x, y) {
    x = x || 0
    y = y || 0
    x = rad(x)
    y = rad(y)
    let c = Math.tan(x).toFixed(9)
    let b = Math.tan(y).toFixed(9)
    return this.add(1, b, c, 1, 0, 0)
  }

  /**
   * Scales the matrix
   * x (number) amount to be scaled, with `1` resulting in no change
   * y (number) #optional amount to scale along the vertical axis. (Otherwise `x` applies to both axes.)
   * cx (number) #optional horizontal origin point from which to scale
   * cy (number) #optional vertical origin point from which to scale
   * Default cx, cy is the middle point of the element.
   */
  scale(x, y, cx, cy) {
    y == null && (y = x)
    ;(cx || cy) && this.translate(cx, cy)
    this.a *= x
    this.b *= x
    this.c *= y
    this.d *= y
    ;(cx || cy) && this.translate(-cx, -cy)
    return this
  }

  /**
   * Returns x coordinate for given point after transformation described by the matrix. See also @Matrix.y
   * x (number)
   * y (number)
   */
  x(x, y) {
    return x * this.a + y * this.c + this.e
  }
  /*
   * Returns y coordinate for given point after transformation described by the matrix. See also @Matrix.x
   * x (number)
   * y (number)
   */
  y(x, y) {
    return x * this.b + y * this.d + this.f
  }
  get(i) {
    return +this[String.fromCharCode(97 + i)].toFixed(4)
  }
  toString() {
    return (
      'matrix(' +
      [
        this.get(0),
        this.get(1),
        this.get(2),
        this.get(3),
        this.get(4),
        this.get(5)
      ].join() +
      ')'
    )
  }

  offset() {
    return [this.e.toFixed(4), this.f.toFixed(4)]
  }

  /**
   * Finds determinant of the given matrix.
   */
  determinant() {
    return this.a * this.d - this.b * this.c
  }
  /**
    * Splits matrix into primitive transformations
    = (object) in format:
    o dx (number) translation by x
    o dy (number) translation by y
    o scalex (number) scale by x
    o scaley (number) scale by y
    o shear (number) shear
    o rotate (number) rotation in deg
    o isSimple (boolean) could it be represented via simple transformations
  */
  split() {
    let out = {}
    // translation
    out.dx = this.e
    out.dy = this.f

    // scale and shear
    let row = [
      [this.a, this.b],
      [this.c, this.d]
    ]
    out.scalex = Math.sqrt(norm(row[0]))
    normalize(row[0])

    out.shear = row[0][0] * row[1][0] + row[0][1] * row[1][1]
    row[1] = [
      row[1][0] - row[0][0] * out.shear,
      row[1][1] - row[0][1] * out.shear
    ]

    out.scaley = Math.sqrt(norm(row[1]))
    normalize(row[1])
    out.shear /= out.scaley

    if (this.determinant() < 0) {
      out.scalex = -out.scalex
    }

    // rotation
    let sin = row[0][1],
      cos = row[1][1]
    if (cos < 0) {
      out.rotate = deg(Math.acos(cos))
      if (sin < 0) {
        out.rotate = 360 - out.rotate
      }
    } else {
      out.rotate = deg(Math.asin(sin))
    }

    out.isSimple =
      !+out.shear.toFixed(9) &&
      (out.scalex.toFixed(9) == out.scaley.toFixed(9) || !out.rotate)
    out.isSuperSimple =
      !+out.shear.toFixed(9) &&
      out.scalex.toFixed(9) == out.scaley.toFixed(9) &&
      !out.rotate
    out.noRotation = !+out.shear.toFixed(9) && !out.rotate
    return out
  }
  /**
   * Returns transform string that represents given matrix
   */
  toTransformString(shorter) {
    let s = shorter || this.split()
    if (!+s.shear.toFixed(9)) {
      s.scalex = +s.scalex.toFixed(4)
      s.scaley = +s.scaley.toFixed(4)
      s.rotate = +s.rotate.toFixed(4)
      return (
        (s.dx || s.dy ? 't' + [+s.dx.toFixed(4), +s.dy.toFixed(4)] : '') +
        (s.rotate ? 'r' + [+s.rotate.toFixed(4), 0, 0] : '') +
        (s.scalex != 1 || s.scaley != 1 ? 's' + [s.scalex, s.scaley, 0, 0] : '')
      )
    } else {
      return (
        'm' +
        [
          this.get(0),
          this.get(1),
          this.get(2),
          this.get(3),
          this.get(4),
          this.get(5)
        ]
      )
    }
  }
}
